/****************************************************************************************/
/*                                                                                      */
/*  Project     :  1D-BRNN                                                              */
/*  Release     :  3.3                                                                  */
/*                                                                                      */
/*  File        :  Train_New_Model.cpp                                                  */
/*  Description :  Train a new BRNN model on a dataset of sequences                     */
/*                                                                                      */
/*  Author(s)   :  Christophe Magnan  (2013) - New generic version                      */
/*              :  Jianlin Cheng      (2003) - New custom version for SCRATCH           */
/*              :  Gianluca Pollastri (2001) - Customized version for SCRATCH           */
/*              :  Paolo Frasconi     (1997) - Initial generic version                  */
/*                                                                                      */
/*  Copyright   :  Institute for Genomics and Bioinformatics                            */
/*                 University of California, Irvine                                     */
/*                                                                                      */
/*  Modified    :  2015/07/01                                                           */
/*                                                                                      */
/****************************************************************************************/

#include "Options.h"
#include "Model.h"

void test_model(Model* BRNN,Dataset* DATA)
{
  for(int i=0;i<DATA->Num_Sequences;i++){
  BRNN->propagate(DATA->Sequences[i]);
  if(i%100==0){ cerr << "." << flush; } }
  cerr << " Done!\n"; DATA->display_confusion();
}

void train_model(Options* OPT,Model* BRNN,Dataset* TRAIN,Dataset* TEST,char* fileout)
{
  // Initial Model Evaluation
  cerr << "Testing initial model on train dataset" << flush; test_model(BRNN,TRAIN);
  cerr << "Testing initial model on test dataset" << flush;  test_model(BRNN,TEST);

  // Training Options
  float epsilon=OPT->LEARN_RATE;
  int num_epochs=OPT->NUM_EPOCHS;
  int num_batchs=OPT->NUM_BATCHS;
  int adap_max_epochs=OPT->ADAP_EPOCHS;
  int adap_reload=OPT->ADAP_RELOAD;
  int shuffle=OPT->SHUFFLE;

  // Initial settings
  int adap_num_epochs=0;
  float previous_error=TRAIN->empirical_risk();
  float current_error=previous_error;
  int num_seq=TRAIN->Num_Sequences;
  int* train_seq=new int[num_seq];
  for(int i=0;i<num_seq;i++){ train_seq[i]=i; }

  // Training iterations
  cerr << "Ready for training the BRNN model...\n\n";
  for(int epoch=0;epoch<num_epochs;epoch++)
  { cerr << "Training epoch " << epoch+1 << " (eps=" << epsilon << ")." << flush;

    // Initializing iteration
    if(shuffle){ int n=num_seq; for(int i=0;i<n;i++){
    int p=rand()%n; int q=rand()%n; int t=train_seq[p];
    train_seq[p]=train_seq[q]; train_seq[q]=t; } }
    float batch_step=(float)num_seq/num_batchs;
    float next_batch=batch_step; int num_pos_batch=0;

    // Back-propagation method
    for(int i=0;i<num_seq;i++)
    { Sequence* seq=TRAIN->Sequences[train_seq[i]];
      BRNN->expectation(seq); num_pos_batch+=seq->Num_Positions;
      if(((i+2)>=next_batch)&&((num_seq-i)>=batch_step))
      { BRNN->maximization(epsilon/(float)num_pos_batch);
        next_batch+=batch_step; num_pos_batch=0; }
      if(i%100==0){ cerr << "." << flush; } }
    BRNN->maximization(epsilon/(float)num_pos_batch);
    current_error=TRAIN->empirical_risk();
    if(current_error != current_error)
    { current_error=2*previous_error; adap_num_epochs=adap_max_epochs; }
    cerr << "\nPrevious error: " << previous_error << " ";
    cerr << "Current error: " << current_error << " -> ";

    // Adaptative Procedure
    if(current_error<previous_error)
    { cerr << "writing new model in \"" << fileout << "\"...\n\n";
      BRNN->write(fileout); previous_error=current_error; adap_num_epochs=0;
      cerr << "Testing new model on train dataset" << flush; test_model(BRNN,TRAIN);
      cerr << "Testing new model on test dataset" << flush;  test_model(BRNN,TEST); }
    else{ adap_num_epochs++;
      if((adap_max_epochs)&&(adap_num_epochs>=adap_max_epochs))
      { adap_num_epochs=0; epsilon*=0.5;
        cerr << "max number of iterations reached !\n";
        cerr << "New learning rate = " << epsilon << "\n\n";
        if(adap_reload){ cerr << "Loading the last model saved...\n\n";
        delete BRNN; BRNN=new Model(); BRNN->load(fileout); BRNN->alloc_backpropagation(); } }
      else{ cerr << adap_num_epochs << " epochs without improvement.\n\n"; } }
  }
}

int main(int argc,char** argv)
{
  // Script Arguments
  char usage[80]=" options_file train_dataset test_dataset output_model\n\n";
  if(argc!=5){ cerr << "\n  Script usage : " << argv[0] << usage; exit(1); }

  // BRNN & Training Options
  cerr << "\n\nReading options to train the BRNN...\n";
  Options* opt=new Options(); opt->load(argv[1],1);
  opt->display(); srand48(opt->SEED); srand(opt->SEED);

  // Train Dataset
  cerr << "Loading training dataset... " << flush;
  Dataset* train=new Dataset(); train->load(argv[2]);
  cerr << "Done!\n" << train->Num_Sequences << " sequences";
  cerr << " (" << train->Num_Positions << " positions)\n\n";

  // Test Dataset
  cerr << "Loading test dataset... " << flush;
  Dataset* test=new Dataset(); test->load(argv[3]);
  cerr << "Done!\n" << test->Num_Sequences << " sequences";
  cerr << " (" << test->Num_Positions << " positions)\n\n";

  // Random Initial Model
  cerr << "Generating random initial model...\n";
  Model* brnn=new Model(); brnn->initialize(opt->FEATURES,
  opt->CLASSES,opt->HIDDEN,opt->CONTEXT_FWD,opt->CONTEXT_BWD,
  opt->OUTPUTS_FWD,opt->OUTPUTS_BWD,opt->HIDDEN_FWD,opt->HIDDEN_BWD);
  cerr << "Writing random initial model in \"" << argv[4];
  cerr << "\"...\n\n"; brnn->write(argv[4]);

  // Data Compatibility
  int c=1; if(train->Num_Features!=test->Num_Features){ c=0; }
  if(train->Num_Classes!=test->Num_Classes){ c=0; }
  if(train->Num_Features!=brnn->Num_Features){ c=0; }
  if(train->Num_Classes!=brnn->Num_Classes){ c=0; }
  if(!c){ cerr << "Incompatible data, cannot proceed\n"; exit(1); }

  // Memory allocation
  train->alloc_propagation(brnn->Num_Outputs_FWD,brnn->Num_Outputs_BWD);
  train->alloc_backpropagation(brnn->Num_Outputs_FWD,brnn->Num_Outputs_BWD);
  test->alloc_propagation(brnn->Num_Outputs_FWD,brnn->Num_Outputs_BWD);
  brnn->alloc_backpropagation();

  // Train Model
  train_model(opt,brnn,train,test,argv[4]); return 1;
}

